package com.thesis.backend.classifiers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.regex.Pattern;

import cc.mallet.fst.CRF;
import cc.mallet.fst.CRFCacheStaleIndicator;
import cc.mallet.fst.CRFOptimizableByBatchLabelLikelihood;
import cc.mallet.fst.CRFOptimizableByLabelLikelihood;
import cc.mallet.fst.CRFTrainerByValueGradients;
import cc.mallet.fst.MultiSegmentationEvaluator;
import cc.mallet.fst.ThreadedOptimizable;
import cc.mallet.fst.TransducerEvaluator;
import cc.mallet.fst.TransducerTrainer;
import cc.mallet.fst.ViterbiWriter;
import cc.mallet.optimize.Optimizable;
import cc.mallet.pipe.Pipe;
import cc.mallet.pipe.PrintInputAndTarget;
import cc.mallet.pipe.SerialPipes;
import cc.mallet.pipe.SimpleTaggerSentence2TokenSequence;
import cc.mallet.pipe.TokenSequence2FeatureVectorSequence;
import cc.mallet.pipe.iterator.LineGroupIterator;
import cc.mallet.pipe.tsf.FeaturesInWindow;
import cc.mallet.pipe.tsf.LexiconMembership;
import cc.mallet.pipe.tsf.RegexMatches;
import cc.mallet.pipe.tsf.TokenFirstPosition;
import cc.mallet.pipe.tsf.TokenTextCharPrefix;
import cc.mallet.types.Instance;
import cc.mallet.types.InstanceList;
import cc.mallet.types.Sequence;
import cc.mallet.types.TokenSequence;

import com.thesis.backend.entity.Paper;
import com.thesis.backend.util.AvgTitlePipe;

/**
 * CRF classifier
 */
public class CRF_Classifier {

	CRF crf;
	String title;
	String method;
	String publisher;
	String year;
	String country;
	String cite_typ;

	/**
	 * @param  path
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public CRF_Classifier(String path) throws FileNotFoundException,
			IOException, ClassNotFoundException {
		loadCRF(path);
	}

	/**
	 * Create, setup and run through MALLET pipes
	 * This is were features are specified
	 * @param dataFileName
	 * @return iData
	 * @throws IOException
	 */
	public InstanceList putThroughPipes(String dataFileName) throws IOException {
		InstanceList iData;
		ArrayList<Pipe> pipes = new ArrayList<Pipe>();

		pipes.add(new SimpleTaggerSentence2TokenSequence());

		pipes.add(new TokenTextCharPrefix("Px1=", 1));
		pipes.add(new TokenTextCharPrefix("Px2=", 2));
		pipes.add(new TokenTextCharPrefix("Px3=", 3));

		pipes.add(new RegexMatches("ALL_DIGITS", Pattern.compile("\\d*")));
		pipes.add(new RegexMatches("CAP_ALL", Pattern.compile("\\p{Lu}*")));
		pipes.add(new RegexMatches("CAP_LETTER", Pattern.compile("^\\p{Lu}.*")));
		pipes.add(new LexiconMembership("STOPWORD", new File(
				"bin/Article_Sys/Data/stoplist_en.txt"), true));
		pipes.add(new TokenFirstPosition("FIRST_TOKEN"));

		// Detect possibly a year
		pipes.add(new RegexMatches("DETCT_YR", Pattern
				.compile("\\19\\d{2}|\\20\\d{2}")));
		// Detect a country
		pipes.add(new LexiconMembership("DETCT_CTRY", new File(
				"bin/Article_Sys/Data/countries.txt"), true));
		// Detect possible method
		pipes.add(new LexiconMembership("DETCT_MTH", new File(
				"bin/Article_Sys/Data/methods2.txt"), true));
		// Detect type of publication
		pipes.add(new LexiconMembership("DETCT_CIT", new File(
				"bin/Article_Sys/Data/cite_typ.txt"), true));
		// Detect possible publisher
		pipes.add(new LexiconMembership("DETCT_PUB", new File(
				"bin/Article_Sys/Data/pub.txt"), true));

		// Feature window
		pipes.add(new FeaturesInWindow("Window-", -1, 2, null, false));

		// Feature uses the average title length found via article
		// Currently just assigns the 7 first tokens in a new text as title.
		pipes.add(new AvgTitlePipe("POS_TITLE", 7));

		pipes.add(new TokenSequence2FeatureVectorSequence());
		// Print labels and input
		pipes.add(new PrintInputAndTarget());

		Pipe pipe = new SerialPipes(pipes);

		iData = new InstanceList(pipe);

		// Run input through all pipes
		iData.addThruPipe(new LineGroupIterator(new BufferedReader(
				new InputStreamReader(new FileInputStream(dataFileName))),
				Pattern.compile("X{8}"), true));

		return iData;
	}

	/**
	 *  Run CRF model on instance
	 *  Returns a list with 6 arraylists(according to labels) which contain the tokens
	 *  
	 * @param instance
	 * @param paper
	 */
	@SuppressWarnings("rawtypes")
	public void runCrf(Instance instance, Paper paper) {
		String titleString = "";
		String methodString = "";
		String countryString = "";
		String pubString = "";
		String citString = "";
		String yearString = "";

		Sequence input = (Sequence) instance.getData();
		TokenSequence sourceTokenSequence = null;
		if (instance.getSource() instanceof TokenSequence)
			sourceTokenSequence = (TokenSequence) instance.getSource();

		Sequence trueOutput = (Sequence) instance.getTarget();
		Sequence predOutput = crf.transduce(input);

		StringBuffer buf = new StringBuffer();

		for (int j = 0; j < trueOutput.size(); j++) {
			if (sourceTokenSequence != null)
				buf.append((sourceTokenSequence.get(j).getText() + ": "));
			buf.append("\n");
			buf.append(trueOutput.get(j).toString() + '/'
					+ predOutput.get(j).toString());

			// FeatureVector fv = (FeatureVector) input.get(j);
			if (trueOutput.get(j).toString().equals("TITLE"))
				titleString += predOutput.get(j).toString();
			else if (trueOutput.get(j).toString().equals("METHOD"))
				methodString += predOutput.get(j).toString();
			else if (trueOutput.get(j).toString().equals("COUNTRY"))
				countryString += predOutput.get(j).toString();
			else if (trueOutput.get(j).toString().equals("YEAR"))
				yearString += predOutput.get(j).toString();
			else if (trueOutput.get(j).toString().equals("PUBLISHER"))
				pubString += predOutput.get(j).toString();
			else if (trueOutput.get(j).toString().equals("CIT_TYP"))
				citString += predOutput.get(j).toString();

		}
		System.out.println(buf.toString());

		paper.setTitle(titleString);
		paper.setMethod(methodString);
		paper.setCountry(countryString);
		paper.setPublisher(pubString);
		paper.setYear(yearString);
		paper.setCite_typ(citString);
	}

	// 
	/**
	 * Train and create CRF_Model
	 * 
	 * @param trainingData
	 * @param testingData
	 * @return crf
	 * @throws FileNotFoundException
	 */
	@SuppressWarnings("unused")
	private CRF trainCrf(InstanceList trainingData, InstanceList testingData)
			throws FileNotFoundException {

		// model
		CRF crf = new CRF(trainingData.getDataAlphabet(),
				trainingData.getTargetAlphabet());

		crf.addFullyConnectedStatesForLabels();

		// initialize model's weights
		crf.setWeightsDimensionAsIn(trainingData, false);

		// Optimize
		CRFOptimizableByLabelLikelihood optLabel = new CRFOptimizableByLabelLikelihood(
				crf, trainingData);

		// CRF trainer
		Optimizable.ByGradientValue[] opts = new Optimizable.ByGradientValue[] { optLabel };
		// L-BFGS as optimizer
		CRFTrainerByValueGradients crfTrainer = new CRFTrainerByValueGradients(
				crf, opts);

		String[] labels = new String[] { "TITLE", "METHOD", "PUBLISHER",
				"YEAR", "CIT_TYP", "COUNTRY" };
		TransducerEvaluator evaluator = new MultiSegmentationEvaluator(
				new InstanceList[] { trainingData, testingData }, new String[] {
						"train", "test" }, labels, labels) {
			@Override
			public boolean precondition(TransducerTrainer tt) {
				// evaluate model every 5 training iterations
				return tt.getIteration() % 10 == 0;
			}
		};
		crfTrainer.addEvaluator(evaluator);

		/* ---- TO PRINT OUT TO FILE ---- */
		ViterbiWriter viterbiWriter = new ViterbiWriter("CRF_OUTPUT",
				new InstanceList[] { trainingData, testingData }, new String[] {
						"train", "test" }) {
			@Override
			public boolean precondition(TransducerTrainer tt) {
				return tt.getIteration() % 48 == 0;
			}
		};
		crfTrainer.addEvaluator(viterbiWriter);

		// Train until convergence
		crfTrainer.setMaxResets(0);
		crfTrainer.train(trainingData, Integer.MAX_VALUE);

		// evaluate
		evaluator.evaluate(crfTrainer);

		return crf;
	}

	public CRF getCrf() {
		return crf;
	}

	/**
	 * Load CRF from file
	 * @param path
	 */
	private void loadCRF(String path) {
		try {
			crf = (CRF) new ObjectInputStream(new FileInputStream(
					new File(path))).readObject();
			System.out.println("CRF MODEL LOADED");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 *  nFoldValidation method. Insert trainingData, and number of folds.
	 * @param trainingData
	 * @param n
	 */
	@SuppressWarnings("unused")
	private void nFoldValidation(InstanceList trainingData, int n) {
		int i = 0;
		InstanceList.CrossValidationIterator cvIter = trainingData
				.crossValidationIterator(n);
		while (cvIter.hasNext()) {
			System.out.println("nValidation Nr:" + (n - i));
			i++;
			InstanceList[] trainTestSplits = cvIter.nextSplit();
			InstanceList trainSplit = trainTestSplits[0];
			InstanceList testSplit = trainTestSplits[1];

			CRF crf = new CRF(trainSplit.getDataAlphabet(),
					trainSplit.getTargetAlphabet());
			crf.addFullyConnectedStatesForLabels();
			crf.setWeightsDimensionAsIn(trainSplit, false);

			CRFOptimizableByBatchLabelLikelihood batchOptLabel = new CRFOptimizableByBatchLabelLikelihood(
					crf, trainSplit, 2);
			ThreadedOptimizable optLabel = new ThreadedOptimizable(
					batchOptLabel, trainSplit, crf.getParameters()
							.getNumFactors(), new CRFCacheStaleIndicator(crf));

			Optimizable.ByGradientValue[] opts = new Optimizable.ByGradientValue[] { optLabel };
			CRFTrainerByValueGradients crfTrainer = new CRFTrainerByValueGradients(
					crf, opts);

			String[] labels = new String[] { "TITLE", "METHOD", "PUBLISHER",
					"YEAR", "CIT_TYP", "COUNTRY" };
			TransducerEvaluator evaluator = new MultiSegmentationEvaluator(
					new InstanceList[] { trainSplit, testSplit }, new String[] {
							"train", "test" }, labels, labels) {
			};

			crfTrainer.setMaxResets(0);
			boolean converged = crfTrainer.train(trainSplit, Integer.MAX_VALUE);

			if (converged) {
				evaluator.evaluate(crfTrainer);
				System.out.println("CRF Trained with Convergence!");
			} else {
				System.out.println("CRF Trained without Convergence!");
			}
			optLabel.shutdown();

		}
	}

	/** 
	 * Save CRF model on disk
	 * @param crf
	 * @param path
	 */
	@SuppressWarnings("unused")
	private void saveCrf(CRF crf, String path) {
		crf.write(new File(path));
	}

}
