package com.thesis.backend.classifiers;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Properties;
import java.util.regex.Pattern;

import cc.mallet.classify.Classifier;
import cc.mallet.classify.ClassifierTrainer;
import cc.mallet.classify.MaxEntTrainer;
import cc.mallet.classify.Trial;
import cc.mallet.classify.evaluate.ConfusionMatrix;
import cc.mallet.fst.CRF;
import cc.mallet.pipe.CharSequence2TokenSequence;
import cc.mallet.pipe.CharSequenceLowercase;
import cc.mallet.pipe.FeatureSequence2FeatureVector;
import cc.mallet.pipe.Input2CharSequence;
import cc.mallet.pipe.Pipe;
import cc.mallet.pipe.PrintInputAndTarget;
import cc.mallet.pipe.SerialPipes;
import cc.mallet.pipe.Target2Label;
import cc.mallet.pipe.TokenSequence2FeatureSequence;
import cc.mallet.pipe.TokenSequenceRemoveStopwords;
import cc.mallet.pipe.iterator.CsvIterator;
import cc.mallet.types.Instance;
import cc.mallet.types.InstanceList;
import cc.mallet.types.Label;
import cc.mallet.types.Labeling;

/**
 * Document classifier , currently set to maximum entropy (MaxEnt)
 * 
 *
 */
public class MaxEnt_Classifier {

	Classifier maxEnt;

	/**
	 * Load model from path
	 * @param path
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public MaxEnt_Classifier(String path) throws FileNotFoundException,
			IOException, ClassNotFoundException {
		loadClassifier(path);
	}

	/**
	  *Create, setup and run through MALLET pipes
	 * This is were features are specified
	 * @param dataFileName
	 * @return iData
	 * @throws UnsupportedEncodingException
	 * @throws FileNotFoundException
	 */
	public InstanceList putThroughPipes(String dataFileName)
			throws UnsupportedEncodingException, FileNotFoundException {
		InstanceList DocList;
		// Begin by importing documents from text to feature sequences
		ArrayList<Pipe> pipeList = new ArrayList<Pipe>();

		// Pipes: lowercase, tokenize, remove stopwords, map to features
		pipeList.add(new Target2Label());
		pipeList.add(new Input2CharSequence());
		pipeList.add(new CharSequenceLowercase());
		pipeList.add(new CharSequence2TokenSequence(Pattern
				.compile("\\p{L}[\\p{L}\\p{P}]+\\p{L}")));
		pipeList.add(new TokenSequenceRemoveStopwords(new File(
				"bin/Article_Sys/Data/stoplist_en.txt"), "UTF-8", false, false,
				false));

		// Sse processed tokens as features
		pipeList.add(new TokenSequence2FeatureSequence());
		pipeList.add(new FeatureSequence2FeatureVector());
		pipeList.add(new PrintInputAndTarget());

		DocList = new InstanceList(new SerialPipes(pipeList));

		// (3, 2, 1) = (data, label, name)
		CsvIterator reader = new CsvIterator(new FileReader(new File(
				dataFileName)), "(\\w+)\\s+(\\w+)\\s+(.*)", 3, 2, 1);

		DocList.addThruPipe(reader);
		return DocList;
	}

	/**
	 * Run Classifier on instance and return predicted label
	 * @param instance
	 * @return predicted label
	 */
	public String runMaxEnt(Instance instance) {
		Labeling labeling = maxEnt.classify(instance).getLabeling();
		Label label = labeling.getBestLabel();
		return label.toString();
	}

	public Classifier getClassifier() {
		return maxEnt;
	}

	/**
	 * Load classifier from path
	 * @param path
	 */
	private void loadClassifier(String path) {
		Properties sp = System.getProperties() ;
        String cfg = sp.getProperty("user.dir") + "\\" + path;
		try {
			maxEnt = (Classifier) new ObjectInputStream(new FileInputStream(
					new File(cfg))).readObject();
				
			System.out.println("CLASSIFIER LOADED");
		} catch (FileNotFoundException e) {
			System.out.println("Filepath " + cfg);
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 	nFoldValidation method. Insert training data, and number of folds.
	 * @param instances
	 * @param n
	 */
	@SuppressWarnings("unused")
	private void nFoldValidation(InstanceList instances, int n) {

		int j = 0;
		InstanceList.CrossValidationIterator cvIter = instances
				.crossValidationIterator(n);
		while (cvIter.hasNext()) {
			System.out.println("nValidation Nr:" + (n - j));
			j++;
			InstanceList[] trainTestSplits = cvIter.nextSplit();
			InstanceList trainSplit = trainTestSplits[0];
			InstanceList testSplit = trainTestSplits[1];

			ClassifierTrainer trainer = new MaxEntTrainer();
			maxEnt = trainer.train(trainSplit);

			Trial t = new Trial(maxEnt, testSplit);
			for (int i = 0; i < maxEnt.getLabelAlphabet().size(); i++) {
				System.out.println("Class '"
						+ maxEnt.getLabelAlphabet().lookupLabel(i) + "': "
						+ " P: " + t.getPrecision(i) + " R: " + t.getRecall(i)
						+ " F1: " + t.getF1(i));
			}
			ConfusionMatrix c = new ConfusionMatrix(t);
			System.out.println("\n");
			System.out.println(c.toString());
		}

	}

	/**
	 * Method for training the classifier. 
	 * @param instances
	 * @return
	 * @throws FileNotFoundException
	 */
	@SuppressWarnings("unused")
	private Classifier trainMaxEnt(InstanceList instances)
			throws FileNotFoundException {
		ClassifierTrainer trainer = new MaxEntTrainer();
		maxEnt = trainer.train(instances);
		maxEnt.print();
		return maxEnt;
	}

	/**
	 * Save classifier on disk
	 * @param path
	 * @throws IOException
	 */
	@SuppressWarnings("unused")
	private void saveClassifier(String path) throws IOException {
		ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(
				new File(path)));
		oos.writeObject(maxEnt);
		oos.close();
	}

}
